/**************************************************************************
Copyright (C) 2002-2004, Marco Barisione <marco.bari@vene.ws>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
**************************************************************************/

#include <windows.h>
#include <tchar.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include "add_path.h"
/* cygwin doesn't define "_TEOF" and "_tfopen" in <tchar.h>. */
#if !defined _TEOF
#   if defined _UNICODE
#       define _TEOF WEOF
#   else
#       define _TEOF EOF
#   endif
#endif
#if !defined _tfopen
#   if defined _UNICODE
#       define _tfopen _wfopen
#   else
#       define _tfopen fopen
#   endif
#endif

/* Using quotes or not should make no difference but some programs (for
   instance cygwin and msys) don't like them. */
/* #define USE_QUOTES */

#define KEY _T("SYSTEM\\CurrentControlSet\\Control\\") \
            _T("Session Manager\\Environment")
#define AUTOEXEC_PATH _T("c:\\autoexec.bat")
#define AUTOEXEC_INTRO "rem ***** The following line was added by " \
                       APP_NAME " *****"
#ifdef USE_QUOTES
#  define AUTOEXEC_CMD "set PATH=%%PATH%%;\"%s\""
#else
#  define AUTOEXEC_CMD "set PATH=%%PATH%%;%s"
#endif

/* A shortcut for the annoying method to get the length of an array. */
#define ARR_LEN(a) (sizeof(a) / sizeof(a[0]))

/* Command line options. */
BOOL param_verbose = FALSE;       /* "/verbose" */
BOOL param_add = TRUE;            /* not "/del" */
BOOL param_show_result = FALSE;   /* "/result"  */
BOOL param_always = FALSE;        /* "/always"  */


const TCHAR help_message[] =
_T(APP_NAME_CAP) _T(" (version ") _T(VERSION) _T(")\n\n")
_T("Adds or removes a directory from the enviroment variable PATH.\n")
_T("The value is stored in the registry under Windows NT and in ")
_T("\"autoexec.bat\" under Windows 95/98/Me.\n\n")
_T("Syntax: \"add_path [options] path\"\n")
_T("Options can be in the form \"/option\", \"-option\" or \"--option\".\n")
_T("/del: \"path\" is not added but deleted form \"autoexec.bat\"")
_T("/the registry;\n")
_T("/verbose: shows many dialogs about what ") _T(APP_NAME) _T(" is doing;\n")
_T("/result: shows the result of the application;\n")
_T("/always: \"path\" is always added, even if it is already present.\n\n")
_T("Note that under Windows 95/98/Me the variable is updated only ")
_T("after a reboot, under Windows NT the variable")
_T("is updated immediately but some application could need to ")
_T("be restarted to see the correct variable.")
_T("\n\n\n")
_T(COPYRIGHT) _T("\n")
LICENSE;


/*************************************************************************
Display "msg" in a message dialog only if "/verbose" was passed on the
command line.
*************************************************************************/
void verb_out(LPCTSTR msg){
    if (param_verbose)
        MessageBox(NULL, msg, _T(APP_NAME), MB_ICONINFORMATION);
}


/*************************************************************************
Display a string in a message dialog only if "/verbose" was passed on the
command line. The string is obtained replacing "%s" in "msg" with "param".
*************************************************************************/
void verb_out_param(LPCTSTR msg, LPCTSTR param){
    LPTSTR str;
    str = malloc((_tcslen(msg) + _tcslen(param) + 1) * sizeof(TCHAR));
    _stprintf(str, msg, param);
    verb_out(str);
    free(str);
}


/*************************************************************************
Display the error "err_msg".
*************************************************************************/
void err_out(LPCTSTR err_msg){
    MessageBox(NULL, err_msg, _T("Error"), MB_ICONSTOP);
}


/*************************************************************************
Display "base_err_msg" followed by the description of the system error
identified by "sys_err".
*************************************************************************/
void err_out_sys(LPCTSTR base_err_msg, LONG sys_err){
    LPTSTR msg_buff;
    LPTSTR err_msg;
    TCHAR err_intro[] = _T("\n\nError returned by the system:\n");
    size_t len;
    DWORD flags = FORMAT_MESSAGE_ALLOCATE_BUFFER |
                  FORMAT_MESSAGE_FROM_SYSTEM |
                  FORMAT_MESSAGE_IGNORE_INSERTS;
    FormatMessage(flags, NULL, sys_err,
                  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                  (LPTSTR)&msg_buff, 0, NULL);
    len = _tcslen(msg_buff) + _tcslen(base_err_msg) + _tcslen(err_intro) + 1;
    err_msg = malloc(len * sizeof(TCHAR));
    _stprintf(err_msg, _T("%s%s%s"), base_err_msg, err_intro, msg_buff);
    err_out(err_msg);
    free(err_msg);
    LocalFree(msg_buff);
}


/*************************************************************************
Works as "strcmp" but the comparison is not case sensitive.
*************************************************************************/
int tcharicmp(LPCTSTR str1, LPCTSTR str2){
    for (; tolower(*str1) == tolower(*str2); ++str1, ++str2)
        if (*str1 == '\0')
            return 0;
    return tolower(*str1) - tolower(*str2);
}


/*************************************************************************
Returns un unquoted copy of "str" (or a copy of "str" if the quotes are
not present). The returned value must be freed with "free".
*************************************************************************/
LPTSTR unquote(LPCTSTR str){
    int last_pos;
    LPTSTR ret;
    size_t new_len;

    last_pos = _tcslen(str) - 1;
    if (last_pos != -1 && str[0] == '"' && str[last_pos] == '"'){
        new_len= (_tcslen(str) - 1);
        ret = malloc(new_len * sizeof(TCHAR));
        lstrcpyn(ret, &str[1], new_len);
    }
    else
        ret = _tcsdup(str);
    return ret;
}


/*************************************************************************
Find "what" in the ";" separated string "str" and returns a pointer to
the first letter of "what" in the string. If "next" is not "NULL" it
points to the first letter after "what" (excluding the trailing ";").
If "what" isn't find the functions returns "NULL".
*************************************************************************/
LPTSTR find_reg_str(LPTSTR str, LPCTSTR what, LPTSTR * next){
    LPTSTR tok_buff;
    LPTSTR curr_tok;
    LPTSTR curr_tok_dup;
    BOOL path_eq;
    TCHAR sh_path1[MAX_PATH], sh_path2[MAX_PATH];
    int pos = -1;
    LPTSTR ret;

    tok_buff = _tcsdup(str);
    curr_tok = _tcstok(tok_buff, _T(";"));
    while (pos == -1 && curr_tok){
        curr_tok_dup = unquote(curr_tok);
        path_eq = GetShortPathName(what, sh_path1, ARR_LEN(sh_path1)) &&
                  GetShortPathName(curr_tok_dup, sh_path2,
                                   ARR_LEN(sh_path2)) &&
                  (tcharicmp(sh_path1, sh_path2) == 0);
        if (path_eq || tcharicmp(what, curr_tok_dup) == 0){
            pos = curr_tok - tok_buff;
        }
        free(curr_tok_dup);
        curr_tok = _tcstok(NULL, _T(";"));
        if (pos != -1 && next){
            if (curr_tok)
                *next = str + (curr_tok - tok_buff);
            else
                *next = str + _tcslen(str);
        }
    }
    free(tok_buff);
    if (pos != -1)
        ret = str + pos;
    else
        ret = NULL;
    return ret;
}


/*************************************************************************
Add "path" to the registry. Return "TRUE" if the path has been added or
was already in the registry, "FALSE" otherwise.
*************************************************************************/
BOOL add_path_reg(LPCTSTR path){
    HKEY key;
    LONG ret;
    DWORD data_size;
    LPTSTR reg_str;
    BOOL found = FALSE;
    BOOL func_ret = TRUE;
    DWORD send_message_result;
    
    ret = RegOpenKeyEx(HKEY_LOCAL_MACHINE, KEY, 0,
                       KEY_WRITE | KEY_READ, &key);
    if (ret != ERROR_SUCCESS){
        err_out_sys(_T("Cannot open registry."), ret);
        return FALSE;
    }

    RegQueryValueEx(key, _T("PATH"), NULL, NULL, NULL, &data_size);
    data_size += _tcslen(path) + 3 ; /* ";" and quotes, "data_size" already
                                        includes '\0'. */
    reg_str = malloc(data_size * sizeof(TCHAR));
    ret = RegQueryValueEx(key, _T("PATH"), NULL, NULL, (LPBYTE)reg_str,
                          &data_size);
    if (ret != ERROR_SUCCESS){
        err_out_sys(_T("Cannot read \"PATH\" key."), ret);
        func_ret = FALSE;
    }
    else{
        if (param_always || !find_reg_str(reg_str, path, NULL)){
            _tcscat(reg_str, _T(";"));
#ifdef USE_QUOTES
            _tcscat(reg_str, _T(";\""));
#endif
            _tcscat(reg_str, path);
#ifdef USE_QUOTES
            _tcscat(reg_str, _T("\""));
#endif
            ret = RegSetValueEx(key, _T("PATH"), 0, REG_EXPAND_SZ,
                                (LPBYTE)reg_str,
                                (_tcslen(reg_str) + 1) * sizeof(TCHAR));
            if (ret != ERROR_SUCCESS){
                err_out_sys(_T("Cannot write \"PATH\" key."), ret);
                func_ret = FALSE;
            }
            /* Is this needed to make the new key avaible? */
            RegFlushKey(key);
            SetLastError(0);
            ret = SendMessageTimeout(HWND_BROADCAST, WM_SETTINGCHANGE, 0,
                                     (LPARAM)"Environment", SMTO_ABORTIFHUNG,
                                     5000, &send_message_result);
            if (ret != ERROR_SUCCESS && GetLastError() != 0)
                err_out_sys(_T("Cannot propagate the new enviroment to ")
                            _T("other processes. The new value will be ")
                            _T("avaible after a reboot."), GetLastError());
        }
        else
            verb_out(_T("Value already exists in the registry."));
    }

    RegCloseKey(key);
    free(reg_str);
    return func_ret;
}


/*************************************************************************
Removes "path" from the registry. Return "TRUE" if the path has been
removed or it wasn't in the registry, "FALSE" otherwise.
*************************************************************************/
BOOL remove_path_reg(LPCTSTR path){
    HKEY key;
    LONG ret;
    DWORD data_size;
    LPTSTR reg_str;
    LPTSTR reg_str2;
    BOOL found = FALSE;
    BOOL func_ret = TRUE;
    LPTSTR next;
    LPTSTR del_part;
    int last_pos;

    ret = RegOpenKeyEx(HKEY_LOCAL_MACHINE, KEY, 0,
                       KEY_WRITE | KEY_READ, &key);
    if (ret != ERROR_SUCCESS){
        err_out_sys(_T("Cannot open registry."), ret);
        return FALSE;
    }

    RegQueryValueEx(key, _T("PATH"), NULL, NULL, NULL, &data_size);
    data_size += _tcslen(path) + 3; /* ";" and quotes,"data_size" already
                                        includes '\0'. */
    reg_str = malloc(data_size * sizeof(TCHAR));
    ret = RegQueryValueEx(key, _T("PATH"), NULL, NULL,
                          (LPBYTE)reg_str, &data_size);
    if (ret != ERROR_SUCCESS){
        err_out_sys(_T("Cannot read \"PATH\" key."), ret);
        func_ret = FALSE;
    }
    else{
        if (del_part = find_reg_str(reg_str, path, &next)){
            reg_str2 = malloc((_tcslen(reg_str) + 1) * sizeof(TCHAR));
            *del_part = '\0';
            _stprintf(reg_str2, _T("%s%s"), reg_str, next);
            last_pos = _tcslen(reg_str2) - 1;
            if (last_pos != -1 && reg_str2[last_pos] == ';')
                reg_str2[last_pos] = '\0';
            ret = RegSetValueEx(key, _T("PATH"), 0, REG_EXPAND_SZ,
                                (LPBYTE)reg_str2,
                                (_tcslen(reg_str2) + 1) * sizeof(TCHAR));
            if (ret != ERROR_SUCCESS){
                err_out_sys(_T("Cannot write \"PATH\" key."), ret);
                func_ret = FALSE;
            }
            free(reg_str2);
        }
        else
            verb_out(_T("Value does not exist in the registry."));
    }

    RegCloseKey(key);
    free(reg_str);
    return func_ret;
    return FALSE;
}


/* Can this program run under Win9x if compiled with unicode support? */
#if !defined _UNICODE

/*************************************************************************
Add "path" to "autoexec.bat". Return "TRUE" if the path has been added or
was already in the file, "FALSE" otherwise.
*************************************************************************/
BOOL add_path_autoexec(LPCTSTR long_path){
    FILE * file;
    LPTSTR path;
    size_t path_size;
    LPTSTR line;
    LPTSTR out_line;
    size_t line_size;
    size_t sz1, sz2;
    LPTSTR autoexec_intro;
    BOOL found;
    BOOL func_ret = TRUE;

    file = _tfopen(AUTOEXEC_PATH, _T("r+"));
    if (!file){
        err_out(_T("Cannot open \"autoexec.bat\"."));
        return FALSE;
    }

    path_size = _tcslen(long_path) + 1;
    path = malloc(path_size * sizeof(TCHAR));
    if (!GetShortPathName(long_path, path, path_size))
        _tcsncpy(path, long_path, path_size);
    sz1 = _tcslen(path) + _tcslen(AUTOEXEC_CMD);
    sz2 = _tcslen(AUTOEXEC_INTRO) + 2 /* '\n' and '\0'. */;
    line_size = sz1 > sz2 ? sz1 : sz2;
    line = malloc(line_size * sizeof(TCHAR));
    out_line = malloc(line_size * sizeof(TCHAR));
    _stprintf(out_line, AUTOEXEC_CMD, path);
    _tcscat(out_line, _T("\n"));
    autoexec_intro = malloc((_tcslen(AUTOEXEC_INTRO) + 2 /* '\0', '\n' */)
                            * sizeof(TCHAR));
    _tcscpy(autoexec_intro, AUTOEXEC_INTRO);
    _tcscat(autoexec_intro, _T("\n"));

    if (!param_always){
        found = FALSE;
        while (!found && _fgetts(line, line_size, file)){
            if (_tcscmp(autoexec_intro, line) == 0){
                _fgetts(line, line_size, file);
                if (_tcscmp(out_line, line) == 0)
                    found = TRUE;
            }
        }
    }
    else
        found = FALSE;

    if (!found){
        if (fseek(file, 0, SEEK_END) != 0 ||
            _fputts(_T("\n"), file) == _TEOF ||
            _fputts(autoexec_intro, file) == _TEOF ||
            _fputts(out_line, file) == _TEOF)
                func_ret = FALSE;
    }
    else
        verb_out(_T("Value already exists in \"autoexec.bat\"."));

    fclose(file);
    free(path);
    free(line);
    free(out_line);
    free(autoexec_intro);

    return func_ret;
}


/*************************************************************************
Removes "path" from "autoexec.bat". Return "TRUE" if the path has been
removed or it wasn't in the file, "FALSE" otherwise.
*************************************************************************/
BOOL remove_path_autoexec(LPTSTR long_path){
    FILE * file;
    LPTSTR path;
    size_t path_size;
    LPTSTR data;
    long file_size;
    LPTSTR expected_text;
    size_t expected_text_size;
    LPTSTR buff;
    size_t buff_size;
    LPTSTR begin_pos;
    LPTSTR final_part;
    size_t fread_ret;
    BOOL func_ret = TRUE;

    file = _tfopen(AUTOEXEC_PATH, _T("rb"));
    if (!file){
        err_out(_T("Cannot open \"autoexec.bat\" for reading."));
        return FALSE;
    }
    fseek(file, 0, SEEK_END);
    file_size = ftell(file);
    data = malloc(file_size + sizeof(TCHAR) /* '\0'. */);
    data[file_size / sizeof(TCHAR)] = '\0';
    fseek(file, 0, SEEK_SET);
    fread_ret = fread(data, file_size, 1, file);
    fclose(file);
    if (fread_ret != 1){
        err_out(_T("Cannot read \"autoexec.bat\"."));
        return FALSE;
    }

    path_size = _tcslen(long_path) + 1;
    path = malloc(path_size * sizeof(TCHAR));
    if (!GetShortPathName(long_path, path, path_size))
        _tcsncpy(path, long_path, path_size);
    buff_size = _tcslen(AUTOEXEC_CMD) + _tcslen(path);
    buff = malloc(buff_size * sizeof(TCHAR));
    expected_text_size = buff_size + _tcslen(AUTOEXEC_INTRO)
                         + 4 /* 2 * '\r\n' */;
    expected_text = malloc(expected_text_size * sizeof(TCHAR));
    _tcscpy(expected_text, AUTOEXEC_INTRO);
    _tcscat(expected_text, _T("\r\n"));
    _stprintf(buff, AUTOEXEC_CMD, path);
    _tcscat(expected_text, buff);
    _tcscat(expected_text, _T("\r\n"));

    begin_pos = _tcsstr(data, expected_text);
    if (begin_pos){
        file = _tfopen(AUTOEXEC_PATH, _T("wb"));
        if (!file){
            err_out(_T("Cannot open \"autoexec.bat\" for writing."));
            func_ret = FALSE;
        }
        else{
            final_part = begin_pos + _tcslen(expected_text);
            if ((fwrite(data, begin_pos - data, 1, file) != 1 &&
                 (begin_pos - data)) || /* "fwrite"fails if the
                                           second argument is 0 */
                (fwrite(final_part, _tcslen(final_part), 1, file) != 1 &&
                 _tcslen(final_part)))
                    func_ret = FALSE;
            fclose(file);
        }
    }
    else
        verb_out(_T("Value does not exist in \"autoexec.bat\"."));

    free(data);
    free(path);
    free(buff);
    free(expected_text);

    return func_ret;
}

#endif /* #if !defined _UNICODE */

/*************************************************************************
Display the help.
*************************************************************************/
void show_help(void){
    MessageBox(NULL, help_message, _T(APP_NAME), MB_ICONINFORMATION);
}


/*************************************************************************
Parse "cmd_line" and returns a string containing the path that need to be
added/removed for "%PATH%". The returned value must be freed with "free".
Return NULL if the path is not needed (e.g. the option "/help" was in the
command line).
*************************************************************************/
LPTSTR parse_cmd_line(LPTSTR cmd_line){
    TCHAR cmd[30];
    LPTSTR buff;
    LPTSTR ret;
    DWORD gfp_ret;
    int i;
    size_t last;

    for (;;){
        while (isspace(*cmd_line))
            ++cmd_line;
        if (*cmd_line == '/' || *cmd_line == '-'){
            ++cmd_line;
            if (*cmd_line == '-' && *(cmd_line - 1) == '-')
                ++cmd_line;
            for (i = 0;
                 !isspace(*cmd_line) && *cmd_line != '\0' &&
                 i < ARR_LEN(cmd) - 1;
                 ++cmd_line, ++i)
                    cmd[i] = *cmd_line;
            cmd[i] = '\0';
            verb_out_param(_T("Found option \"%s\"."), cmd);
            if (_tcscmp(cmd, _T("verbose")) == 0){
                param_verbose = TRUE;
                param_show_result = TRUE;
            }
            else if (_tcscmp(cmd, _T("del")) == 0){
                param_add = FALSE;
            }
            else if (_tcscmp(cmd, _T("result")) == 0){
                param_show_result = TRUE;
            }
            else if (_tcscmp(cmd, _T("always")) == 0){
                param_always = TRUE;
            }
            else if (_tcscmp(cmd, _T("help")) == 0 ||
                     _tcscmp(cmd, _T("h")) == 0 ||
                     _tcscmp(cmd, _T("?")) == 0 ||
                     _tcscmp(cmd, _T("version")) == 0 ||
                     _tcscmp(cmd, _T("license")) == 0){
                show_help();
                return NULL;
            }
            else{
                TCHAR msg_err[60];
                _stprintf(msg_err, _T("Unknown option: \"%s\"."), cmd);
                err_out(msg_err);
                show_help();
                return NULL;
            }
        }
        else if (*cmd_line == '\0'){
            err_out(_T("Path missing."));
            show_help();
            return FALSE;
        }
        else{
            if (*cmd_line == '"')
                ++cmd_line;
            last = _tcslen(cmd_line) - 1;
            if (cmd_line[last] == '"')
                cmd_line[last] = '\0';
            buff = malloc((_tcslen(cmd_line) + 1) * sizeof(TCHAR));
            _tcscpy(buff, cmd_line);
            if (param_always && !param_add)
                verb_out(_T("\"/always\" is unused with \"/del\"."));
            ret = malloc(MAX_PATH * sizeof(TCHAR));
            gfp_ret = GetFullPathName(buff, MAX_PATH, ret, NULL);
            if (gfp_ret > MAX_PATH){
                free(ret);
                ret = malloc((gfp_ret + 1) * sizeof(TCHAR));
                GetFullPathName(buff, gfp_ret + 1, ret, NULL);
            }
            free(buff);
            return ret;
        }
    }
}


/*************************************************************************
Entry point of the application.
*************************************************************************/
INT APIENTRY _tWinMain(HINSTANCE hInstance,
                       HINSTANCE hPrevInstance,
                       LPTSTR    lpCmdLine,
                       INT       nCmdShow){
    OSVERSIONINFO winver;
    LPTSTR path;
    BOOL ret;
    path = parse_cmd_line(lpCmdLine);
    if (path){
        winver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
        GetVersionEx(&winver);
        if (winver.dwPlatformId == 1){
#if !defined _UNICODE
            if (param_add){
                verb_out_param(_T("You are using Windows 95/98/Me.\n")
                               _T("Adding \"%s\" to \"autoexec.bat\"."),
                               path);
                ret = add_path_autoexec(path);
            }
            else{
                verb_out_param(_T("You are using Windows 95/98/Me.\n")
                               _T("Removing \"%s\" from \"autoexec.bat\"."),
                               path);
                ret = remove_path_autoexec(path);
            }
#else
            err_out(_T("The Unicode version of this program cannot run ")
                    _T("under Windows 95/98/Me."));
            ret = FALSE;
#endif /* #if !defined _UNICODE */
        }
        else{
            if (param_add){
                verb_out_param(_T("You are using Windows NT.\n")
                               _T("Adding \"%s\" to the registry."),
                               path);
                ret = add_path_reg(path);
            }
            else{
                verb_out_param(_T("You are using Windows NT.\n")
                               _T("Removing \"%s\" from the registry."),
                               path);
                ret = remove_path_reg(path);
            }
        }
    }
    if (param_show_result){
        if (ret)
            MessageBox(NULL, _T("%PATH% has been updated correctly."),
                       _T(APP_NAME), MB_ICONINFORMATION);
        else
            MessageBox(NULL, _T("%PATH% has not been updated correctly."),
                       _T(APP_NAME), MB_ICONEXCLAMATION);
    }
    free(path);
    return ret ? EXIT_SUCCESS : EXIT_FAILURE;
}
